package main

import "flag"
import "fmt"
import "log"
import "os"
import "runtime/pprof"
import "shardkv"

var cpuprofile = flag.String("cpuprofile", "", "write cpu profile to this file")
var memprofile = flag.String("memprofile", "", "write memory profile to this file")

var scenario = flag.String("scenario", "", "name of the benchmark scenario to run")
var unreliable = flag.Bool("unreliable", false, "unreliable RPC?")

var ng = flag.Int("ng", 3, "number of groups")
var np = flag.Int("np", 3, "number of peers per group")
var ns = flag.Int("ns", 10, "number of shards")
var nc = flag.Int("nc", 3, "number of clients")
var rps = flag.Int("rps", 1, "requests per second, per client")
var dur = flag.Int("dur", 20, "how many seconds to run the benchmark")

var p = flag.Int("p", 0, "persistence?")
var m = flag.Int("m", 0, "mencius?")
var c = flag.Int("c", 0, "caching?")


func main() {
  flag.Parse()
  
  fmt.Printf("Running with %v clients at %v req/s for %v seconds.\n",
    *nc, *rps, *dur)
  fmt.Printf("Persistence? %v\n", *p)
  fmt.Printf("Mencius? %v\n", *m)
  fmt.Printf("Caching? %v\n", *c)
  
  if *cpuprofile != "" {
    f, err := os.Create(*cpuprofile)
    defer f.Close()
    if err != nil {
        log.Fatal(err)
    }
    pprof.StartCPUProfile(f)
    defer pprof.StopCPUProfile()
  }
  
  shardkv.Benchmark(*scenario, *unreliable, *ng, *np, *ns, *nc, *rps, *dur, *p, *m, *c)

  if *memprofile != "" {
    f, err := os.Create(*memprofile)
    defer f.Close()
    if err != nil {
      log.Fatal(err)
    }
    pprof.WriteHeapProfile(f)
  }
}
