package main

import "os"
import "runtime/pprof"
import "log"
import "shardkv"
import "shardmaster"
import "runtime"
import "strconv"
import "time"
import "fmt"
import "math/rand"
import "runtime/debug"
import "flag"

var memprofile = flag.String("memprofile", "", "write memory profile to this file")

func main() {

  runtime.GOMAXPROCS(8)

	flag.Parse()

  shardkv.NShards = 10000

	dostuff()
	
  var m runtime.MemStats

  runtime.GC()
	runtime.ReadMemStats(&m)
	log.Printf("after returning: numgc = %d bytes = %d frees = %d total = %d footprint = %d released = %d", m.NumGC, m.HeapAlloc, m.Frees, m.TotalAlloc, m.Sys, m.HeapReleased)

	time.Sleep(1 * time.Second)
	
  runtime.GC()
  var m1 runtime.MemStats
  runtime.ReadMemStats(&m1)

	debug.FreeOSMemory()

	runtime.ReadMemStats(&m)
	log.Printf("after free os mem: numgc = %d bytes = %d frees = %d total = %d footprint = %d released = %d", m.NumGC, m.HeapAlloc, m.Frees, m.TotalAlloc, m.Sys, m.HeapReleased)
}

func dostuff() {
  smh, gids, ha, _, clean := shardkv.Setup("big-main", false)
  defer clean()

  fmt.Printf("Test: Lots of data ...\n")

  mck := shardmaster.MakeClerk(smh)
  mck.Join(gids[0], ha[0])
  mck.Join(gids[1], ha[1])

	time.Sleep(1 * time.Second)

  ck := shardkv.MakeClerk(shardkv.NShards, smh)

	// get the current memory, putting keys after this shouldn't change
	// memory much

  var m runtime.MemStats

	runtime.ReadMemStats(&m)
	log.Printf("before anything: numgc = %d bytes = %d frees = %d total = %d footprint = %d released = %d", m.NumGC, m.HeapAlloc, m.Frees, m.TotalAlloc, m.Sys, m.HeapReleased)
	
	// put a bunch of other keys
	putStuff(ck)

	time.Sleep(10 * time.Minute)

  runtime.GC()
	runtime.ReadMemStats(&m)
	log.Printf("after puts: numgc = %d bytes = %d frees = %d total = %d footprint = %d released = %d", m.NumGC, m.HeapAlloc, m.Frees, m.TotalAlloc, m.Sys, m.HeapReleased)

  if *memprofile != "" {
    f, err := os.Create(*memprofile)
    defer f.Close()
    if err != nil {
      log.Fatal(err)
    }
    pprof.WriteHeapProfile(f)
  }	

	clean()

  runtime.GC()
	runtime.ReadMemStats(&m)
	log.Printf("after clean: numgc = %d bytes = %d frees = %d total = %d footprint = %d released = %d", m.NumGC, m.HeapAlloc, m.Frees, m.TotalAlloc, m.Sys, m.HeapReleased)

}

func putStuff(ck *shardkv.Clerk) {

  var m runtime.MemStats

	nOtherKeys := 1000000000
	valSize := 100

  by := 1000

	done := 0
  for k := 0; k < nOtherKeys/by; k++ {

    doneChannel := make(chan bool)
    for i := 0; i < by; i++ {
      go func() {
        key := strconv.Itoa(rand.Int())
        big := make([]byte, valSize)
        for j := 0; j < len(big); j++ {
          big[j] = byte('a' + rand.Int() % 26)
        }
        ck.Put(key, string(big))
        doneChannel<-true
      }()
    }
    
    for done < (k+1)*by {
      select {
      case <-doneChannel:
        done++
      }
      if done % by == 0 {
        fmt.Println("done with putting", done)
        runtime.GC()
        runtime.ReadMemStats(&m)
        log.Printf("after clean: numgc = %d bytes = %d frees = %d total = %d footprint = %d released = %d", m.NumGC, m.HeapAlloc, m.Frees, m.TotalAlloc, m.Sys, m.HeapReleased)
      }
    }

    close(doneChannel)
  }

}
