package paxos

import "testing"
import "runtime"
import "strconv"
import "os"
import "time"
import "fmt"
import "math/rand"
import "sync"

var mu sync.Mutex = sync.Mutex{}
var persist bool = true
var mencius bool = false

func port(tag string, host int) string {
  s := "/var/tmp/824-"
  s += strconv.Itoa(os.Getuid()) + "/"
  os.Mkdir(s, 0777)
  s += "px-"
  s += strconv.Itoa(os.Getpid()) + "-"
  s += tag + "-"
  s += strconv.Itoa(host)
  return s
}

func safeStart(pxa []*Paxos, i int, seq int, v interface{}) {
	mu.Lock()
	pxa[i].Start(seq, v)
	mu.Unlock()
}

func ndecided(t *testing.T, pxa []*Paxos, seq int) int {
  count := 0
  var v interface{}
  for i := 0; i < len(pxa); i++ {
		mu.Lock()
		here := pxa[i] != nil
		mu.Unlock()
    if here {
			mu.Lock()
      decided, v1 := pxa[i].Status(seq)
      //fmt.Printf("Checker retreived %d from %d in slot %d\n", v1, i, seq)
			mu.Unlock()
      if decided {
        if count > 0 && v != v1 {
          fmt.Printf("v is from %v, v1 from %v\n", i-1, i)
          t.Fatalf("decided values do not match; seq=%v i=%v v=%v v1=%v",
            seq, i, v, v1)
        }
        count++
        v = v1
      }
    }
  }
  return count
}

// randomly kills processes, and randomly deletes their disks
// at most a minority of processes will ever lose their disk
// all processes are guaranteed to be alive after returning
func crashRandomly(unreliable bool, pxa []*Paxos, doneChannel chan bool) {
	diskless := 0
	dead := make(map[int]bool)
	for {
		select {
		case <-doneChannel:
			mu.Lock()
			for i := range dead {
				if dead[i] {
					pxh := pxa[i].peers // is it possible to get the peers back from a dead paxos? what if disk was lost?
					pxa[i] = Make(mencius, persist, pxh, i, nil)
				}
			}
			mu.Unlock()
			doneChannel <- true
			return
		default:
			for i := 0; i < len(pxa); i++ {
				mu.Lock()
				if pxa[i] != nil {
					pxh := pxa[i].peers
					if rand.Intn(10) < 1 {
						fmt.Println("killing", i)
						if rand.Intn(10) < 1 && diskless < len(pxa) - (len(pxa)/2 + 1) {
							diskless += 1
							pxa[i].Kill(false)
							dead[i] = true
						} else {
							pxa[i].Kill(true)
							dead[i] = true
						}
						time.Sleep(10 * time.Millisecond)
					}
					if dead[i] && rand.Intn(10) < 5 {
						fmt.Println("reviving", i)
						dead[i] = false
						seq := pxa[i].Max()
						pxa[i] = Make(mencius, persist, pxh, i, nil)
						if unreliable {
							pxa[i].unreliable = true
						}
						for s := 0; s <= seq; s++ {
							pxa[i].Start(s, s) // this is a hack
						}
						time.Sleep(10 * time.Millisecond)
					}
				}
				mu.Unlock()
			}
			time.Sleep(10 * time.Millisecond)
		}
	}
}

func safeClean(pxa []*Paxos, doneChannel chan bool) {
		doneChannel <- true
		<-doneChannel
		cleanup(pxa)
}

func waitn(t *testing.T, pxa[]*Paxos, seq int, wanted int) {
  to := 10 * time.Millisecond
  for iters := 0; iters < 30; iters++ {
    if ndecided(t, pxa, seq) >= wanted {
      break
    }
    time.Sleep(to)
    if to < time.Second {
      to *= 2
    }
  }
  nd := ndecided(t, pxa, seq)
  if nd < wanted {
    t.Fatalf("too few decided; seq=%v ndecided=%v wanted=%v", seq, nd, wanted)
  }
}

func waitmajority(t *testing.T, pxa[]*Paxos, seq int) {
  waitn(t, pxa, seq, (len(pxa) / 2) + 1)
}

func checkmax(t *testing.T, pxa[]*Paxos, seq int, max int) {
  time.Sleep(3 * time.Second)
  nd := ndecided(t, pxa, seq)
  if nd > max {
    t.Fatalf("too many decided; seq=%v ndecided=%v max=%v", seq, nd, max)
  }
}

func cleanup(pxa []*Paxos) {
	mu.Lock()
  for i := 0; i < len(pxa); i++ {
    if pxa[i] != nil {
      pxa[i].Kill(false)
    }
  }
	mu.Unlock()
}

func TestSpeed(t *testing.T) {
  runtime.GOMAXPROCS(4)

  const npaxos = 3
  var pxa []*Paxos = make([]*Paxos, npaxos)
  var pxh []string = make([]string, npaxos)
  defer cleanup(pxa)

  for i := 0; i < npaxos; i++ {
    pxh[i] = port("time", i)
  }
  for i := 0; i < npaxos; i++ {
    pxa[i] = Make(mencius, persist, pxh, i, nil)
  }

  t0 := time.Now()

  for i := 0; i < 20; i++ {
    pxa[0].Start(pxa[0].GetNextSlot(), "x")
    waitn(t, pxa, i, npaxos)
  }

  d := time.Since(t0)
  fmt.Printf("20 agreements %v seconds\n", d.Seconds())
}

func TestBasic(t *testing.T) {
  runtime.GOMAXPROCS(4)

  const npaxos = 3
  var pxa []*Paxos = make([]*Paxos, npaxos)
  var pxh []string = make([]string, npaxos)
  defer cleanup(pxa)

  for i := 0; i < npaxos; i++ {
    pxh[i] = port("basic", i)
  }
  for i := 0; i < npaxos; i++ {
    pxa[i] = Make(mencius, persist, pxh, i, nil)
  }

  fmt.Printf("Test: Single proposer ...\n")

  pxa[0].Start(0, "hello")
  waitn(t, pxa, 0, npaxos)

  fmt.Printf("  ... Passed\n")

  fmt.Printf("Test: Many proposers, same value ...\n")

  for i := 0; i < npaxos; i++ {
    pxa[i].Start(1, 77)
  }
  waitn(t, pxa, 1, npaxos)

  fmt.Printf("  ... Passed\n")

  fmt.Printf("Test: Many proposers, different values ...\n")

  pxa[0].Start(2, 100)
  pxa[1].Start(2, 101)
  pxa[2].Start(2, 102)
  waitn(t, pxa, 2, npaxos)

  fmt.Printf("  ... Passed\n")

  fmt.Printf("Test: Out-of-order instances ...\n")

  pxa[0].Start(7, 700)
  pxa[0].Start(6, 600)
  pxa[1].Start(5, 500)
  waitn(t, pxa, 7, npaxos)
  pxa[0].Start(4, 400)
  pxa[1].Start(3, 300)
  waitn(t, pxa, 6, npaxos)
  waitn(t, pxa, 5, npaxos)
  waitn(t, pxa, 4, npaxos)
  waitn(t, pxa, 3, npaxos)

  if pxa[0].Max() != 7 {
    t.Fatalf("wrong Max()")
  }

  fmt.Printf("  ... Passed\n")
}

func TestDeaf(t *testing.T) {
  runtime.GOMAXPROCS(4)

  const npaxos = 5
  var pxa []*Paxos = make([]*Paxos, npaxos)
  var pxh []string = make([]string, npaxos)
  defer cleanup(pxa)

  for i := 0; i < npaxos; i++ {
    pxh[i] = port("deaf", i)
  }
  for i := 0; i < npaxos; i++ {
    pxa[i] = Make(mencius, persist, pxh, i, nil)
  }

  fmt.Printf("Test: Deaf proposer ...\n")

  pxa[0].Start(0, "hello")
  waitn(t, pxa, 0, npaxos)

  os.Remove(pxh[0])
  os.Remove(pxh[npaxos-1])

  pxa[1].Start(1, "goodbye")
	
  waitmajority(t, pxa, 1)
  time.Sleep(1 * time.Second)

  if ndecided(t, pxa, 1) != npaxos - 2 {
    t.Fatalf("a deaf peer heard about a decision")
  }

  pxa[0].Start(1, "xxx")
  waitn(t, pxa, 1, npaxos-1)
  time.Sleep(1 * time.Second)
  if ndecided(t, pxa, 1) != npaxos - 1 {
    t.Fatalf("a deaf peer heard about a decision")
  }

  pxa[npaxos-1].Start(1, "yyy")
  waitn(t, pxa, 1, npaxos)

  fmt.Printf("  ... Passed\n")
}

func TestForget(t *testing.T) {
  runtime.GOMAXPROCS(4)

  const npaxos = 6
  var pxa []*Paxos = make([]*Paxos, npaxos)
  var pxh []string = make([]string, npaxos)
  defer cleanup(pxa)
  
  for i := 0; i < npaxos; i++ {
    pxh[i] = port("gc", i)
  }
  for i := 0; i < npaxos; i++ {
    pxa[i] = Make(mencius, persist, pxh, i, nil)
  }

  fmt.Printf("Test: Forgetting ...\n")

  // initial Min() correct?
  for i := 0; i < npaxos; i++ {
    m := pxa[i].Min()
    if m > 0 {
      t.Fatalf("wrong initial Min() %v", m)
    }
  }

  pxa[0].Start(0, "00")
  pxa[1].Start(1, "11")
  pxa[2].Start(2, "22")
  pxa[0].Start(6, "66")
  pxa[1].Start(7, "77")

  waitn(t, pxa, 0, npaxos)

  // Min() correct?
  for i := 0; i < npaxos; i++ {
    m := pxa[i].Min()
    if m != 0 {
      t.Fatalf("wrong Min() %v; expected 0", m)
    }
  }

  waitn(t, pxa, 1, npaxos)

  // Min() correct?
  for i := 0; i < npaxos; i++ {
    m := pxa[i].Min()
    if m != 0 {
      t.Fatalf("wrong Min() %v; expected 0", m)
    }
  }

  // everyone Done() -> Min() changes?
  for i := 0; i < npaxos; i++ {
    pxa[i].Done(0)
  }
  for i := 1; i < npaxos; i++ {
    pxa[i].Done(1)
  }
  for i := 0; i < npaxos; i++ {
    pxa[i].Start(8 + i, "xx")
  }
  allok := false
  for iters := 0; iters < 12; iters++ {
    allok = true
    for i := 0; i < npaxos; i++ {
      s := pxa[i].Min()
      if s != 1 {
        allok = false
      }
    }
    if allok {
      break
    }
    time.Sleep(1 * time.Second)
  }
  if allok != true {
    t.Fatalf("Min() did not advance after Done()")
  }

  fmt.Printf("  ... Passed\n")
}

func TestManyForget(t *testing.T) {
  runtime.GOMAXPROCS(4)

  const npaxos = 3
  var pxa []*Paxos = make([]*Paxos, npaxos)
  var pxh []string = make([]string, npaxos)
  defer cleanup(pxa)
  
  for i := 0; i < npaxos; i++ {
    pxh[i] = port("manygc", i)
  }
  for i := 0; i < npaxos; i++ {
    pxa[i] = Make(mencius, persist, pxh, i, nil)
    pxa[i].unreliable = true
  }

  fmt.Printf("Test: Lots of forgetting ...\n")

  const maxseq = 20
  done := false

  go func() {
    na := rand.Perm(maxseq)
    for i := 0; i < len(na); i++ {
      seq := na[i]
      j := (rand.Int() % npaxos)
      v := rand.Int() 
      pxa[j].Start(seq, v)
      runtime.Gosched()
    }
  }()

  go func() {
    for done == false {
      seq := (rand.Int() % maxseq)
      i := (rand.Int() % npaxos)
      if seq >= pxa[i].Min() {
        decided, _ := pxa[i].Status(seq)
        if decided {
          pxa[i].Done(seq)
        }
      }
      runtime.Gosched()
    }
  }()

  time.Sleep(5 * time.Second)
  done = true
  for i := 0; i < npaxos; i++ {
    pxa[i].unreliable = false
  }
  time.Sleep(2 * time.Second)

  for seq := 0; seq < maxseq; seq++ {
    for i := 0; i < npaxos; i++ {
      if seq >= pxa[i].Min() {
        pxa[i].Status(seq)
      }
    }
  }

  fmt.Printf("  ... Passed\n")
}

//
// does paxos forgetting actually free the memory?
//
func TestForgetMem(t *testing.T) {
	ForgetMem(t, false)
}

func ForgetMem(t *testing.T, crashes bool) {
  runtime.GOMAXPROCS(4)

  fmt.Printf("Test: Paxos frees forgotten instance memory ...\n")

  const npaxos = 3
  var pxa []*Paxos = make([]*Paxos, npaxos)
  var pxh []string = make([]string, npaxos)
  defer cleanup(pxa)
  
  for i := 0; i < npaxos; i++ {
    pxh[i] = port("gcmem", i)
  }
  for i := 0; i < npaxos; i++ {
    pxa[i] = Make(mencius, persist, pxh, i, nil)
  }

  pxa[0].Start(0, "x")
  waitn(t, pxa, 0, npaxos)

  runtime.GC()
  var m0 runtime.MemStats
  runtime.ReadMemStats(&m0)
  // m0.Alloc about a megabyte

  for i := 1; i <= 10; i++ {
    big := make([]byte, 1000000)
    for j := 0; j < len(big); j++ {
      big[j] = byte('a' + rand.Int() % 26)
    }
    pxa[0].Start(i, string(big))
    waitn(t, pxa, i, npaxos)
  }

  runtime.GC()
  var m1 runtime.MemStats
  runtime.ReadMemStats(&m1)
  // m1.Alloc about 90 megabytes

	if crashes {
		pxa[0].Kill(false)
		pxa[1].Kill(true)
		pxa[2].Kill(true)

		pxa[0] = Make(mencius, persist, pxh, 0, nil)
		pxa[1] = Make(mencius, persist, pxh, 1, nil)
		pxa[2] = Make(mencius, persist, pxh, 2, nil)
		time.Sleep(1 * time.Second)

	}

  for i := 0; i < npaxos; i++ {
    pxa[i].Done(10)
  }
  for i := 0; i < npaxos; i++ {
    pxa[i].Start(11 + i, "z")
  }
  time.Sleep(1 * time.Second)
  for i := 0; i < npaxos; i++ {
    if pxa[i].Min() != 11 {
      t.Fatalf("expected Min() %v, got %v\n", 11, pxa[i].Min())
    }
  }

  runtime.GC()
  var m2 runtime.MemStats
  runtime.ReadMemStats(&m2)
  // m2.Alloc about 10 megabytes

  if m2.Alloc > (m1.Alloc / 2) {
    t.Fatalf("memory use did not shrink enough")
  }

  fmt.Printf("  ... Passed\n")
}

func TestRPCCount(t *testing.T) {
  runtime.GOMAXPROCS(4)

  fmt.Printf("Test: RPC counts aren't too high ...\n")

  const npaxos = 3
  var pxa []*Paxos = make([]*Paxos, npaxos)
  var pxh []string = make([]string, npaxos)
  defer cleanup(pxa)

  for i := 0; i < npaxos; i++ {
    pxh[i] = port("count", i)
  }
  for i := 0; i < npaxos; i++ {
    pxa[i] = Make(mencius, persist, pxh, i, nil)
  }

  ninst1 := 5
  seq := 0
  for i := 0; i < ninst1; i++ {
    pxa[0].Start(seq, "x")
    waitn(t, pxa, seq, npaxos)
    seq++
  }

  time.Sleep(2 * time.Second)

  total1 := 0
  for j := 0; j < npaxos; j++ {
    total1 += pxa[j].rpcCount
  }

  // per agreement:
  // 3 prepares
  // 3 accepts
  // 3 decides
  expected1 := ninst1 * npaxos * npaxos
  if total1 > expected1 {
    t.Fatalf("too many RPCs for serial Start()s; %v instances, got %v, expected %v",
      ninst1, total1, expected1)
  }
  fmt.Printf("Test: RPC counts aren't too high ...\n")

  ninst2 := 5
  for i := 0; i < ninst2; i++ {
    for j := 0; j < npaxos; j++ {
      //go pxa[j].Start(seq, j + (i * 10))

      go pxa[j].Start(pxa[j].GetNextSlot(), j + (i * 10))
    }
    //waitn(t, pxa, seq, npaxos)
    seq++
  }

  time.Sleep(2 * time.Second)

  total2 := 0
  for j := 0; j < npaxos; j++ {
    total2 += pxa[j].rpcCount
  }
  total2 -= total1

  // worst case per agreement:
  // Proposer 1: 3 prep, 3 acc, 3 decides.
  // Proposer 2: 3 prep, 3 acc, 3 prep, 3 acc, 3 decides.
  // Proposer 3: 3 prep, 3 acc, 3 prep, 3 acc, 3 prep, 3 acc, 3 decides.
  expected2 := ninst2 * npaxos * 15
  if total2 > expected2 {
    t.Fatalf("too many RPCs for concurrent Start()s; %v instances, got %v, expected %v",
      ninst2, total2, expected2)
  } else {
    fmt.Printf("RPC count : %d    original expected: %d\n", total2, expected2)
  }

  fmt.Printf("  ... Passed\n")
}

//
// many agreements (without failures)
//
func TestMany(t *testing.T) {
	Many(t, false)
}

func Many(t *testing.T, crashes bool) {
  runtime.GOMAXPROCS(4)

	if crashes {
		fmt.Printf("Test: Many instances with crashes ...\n")
	} else {
		fmt.Printf("Test: Many instances ...\n")
	}


  const npaxos = 3
  var pxa []*Paxos = make([]*Paxos, npaxos)
  var pxh []string = make([]string, npaxos)

  for i := 0; i < npaxos; i++ {
		if crashes {
			pxh[i] = port("manycrashes", i)
		} else {
			pxh[i] = port("many", i)
		}
  }
  for i := 0; i < npaxos; i++ {
    pxa[i] = Make(mencius, persist, pxh, i, nil)
    pxa[i].Start(0, 0)
  }

	if crashes {
		doneChannel := make(chan bool)
		go crashRandomly(false, pxa, doneChannel)
		defer safeClean(pxa, doneChannel)
	} else {
		defer cleanup(pxa)
	}

  const ninst = 50
  for seq := 1; seq < ninst; seq++ {
    // only 5 active instances, to limit the
    // number of file descriptors.
    for seq >= 5 && ndecided(t, pxa, seq - 5) < npaxos {
      time.Sleep(20 * time.Millisecond)
    }
    for i := 0; i < npaxos; i++ {
			mu.Lock()
      pxa[i].Start(seq, (seq * 10) + i)
			mu.Unlock()
    }
  }

  for {
    done := true
    for seq := 1; seq < ninst; seq++ {
			if ndecided(t, pxa, seq) < npaxos {
				done = false
			}
    }
    if done {
      break
    }
    time.Sleep(100 * time.Millisecond)
  }

  fmt.Printf("  ... Passed\n")
}

//
// a peer starts up, with proposal, after others decide.
// then another peer starts, without a proposal.
// 
func TestOld(t *testing.T) {
  runtime.GOMAXPROCS(4)

  fmt.Printf("Test: Minority proposal ignored ...\n")

  const npaxos = 5
  var pxa []*Paxos = make([]*Paxos, npaxos)
  var pxh []string = make([]string, npaxos)
  defer cleanup(pxa)

  for i := 0; i < npaxos; i++ {
    pxh[i] = port("old", i)
  }

  pxa[1] = Make(mencius, persist, pxh, 1, nil)
  pxa[2] = Make(mencius, persist, pxh, 2, nil)
  pxa[3] = Make(mencius, persist, pxh, 3, nil)
  pxa[1].Start(1, 111)

  waitmajority(t, pxa, 1)

  pxa[0] = Make(mencius, persist, pxh, 0, nil)
  pxa[0].Start(1, 222)

  waitn(t, pxa, 1, 4)

  if false {
    pxa[4] = Make(mencius, persist, pxh, 4, nil)
    waitn(t, pxa, 1, npaxos)
  }

  fmt.Printf("  ... Passed\n")
}

//
// many agreements, with unreliable RPC
//
func TestManyUnreliable(t *testing.T) {
	ManyUnreliable(t, false)
}

func ManyUnreliable(t *testing.T, crashes bool) {
  runtime.GOMAXPROCS(4)

	if crashes {
		fmt.Printf("Test: Many instances, unreliable RPC and crashes...\n")
	} else {
		fmt.Printf("Test: Many instances, unreliable RPC ...\n")
	}

  const npaxos = 3
  var pxa []*Paxos = make([]*Paxos, npaxos)
  var pxh []string = make([]string, npaxos)

  for i := 0; i < npaxos; i++ {
		if crashes {
			pxh[i] = port("manyuncrashes", i)
		} else {
			pxh[i] = port("manyun", i)
		}
  }
	
	doneChannel := make(chan bool)
	if crashes {
		defer safeClean(pxa, doneChannel)
	} else {
		defer cleanup(pxa)
	}

  for i := 0; i < npaxos; i++ {
    pxa[i] = Make(mencius, persist, pxh, i, nil)
    pxa[i].unreliable = true
    pxa[i].Start(0, 0)
  }

	if crashes {
		go crashRandomly(true, pxa, doneChannel)
	}

  const ninst = 50
  for seq := 1; seq < ninst; seq++ {
    // only 3 active instances, to limit the
    // number of file descriptors.
    for seq >= 3 && ndecided(t, pxa, seq - 3) < npaxos {
      time.Sleep(20 * time.Millisecond)
    }
    for i := 0; i < npaxos; i++ {
			mu.Lock()
      pxa[i].Start(seq, (seq * 10) + i)
			mu.Unlock()
    }
  }

  for {
    done := true
    for seq := 1; seq < ninst; seq++ {
      if ndecided(t, pxa, seq) < npaxos {
        done = false
      }
    }
    if done {
      break
    }
    time.Sleep(100 * time.Millisecond)
  }
  
  fmt.Printf("  ... Passed\n")
}

func pp(tag string, src int, dst int) string {
  s := "/var/tmp/824-"
  s += strconv.Itoa(os.Getuid()) + "/"
  s += "px-" + tag + "-"
  s += strconv.Itoa(os.Getpid()) + "-"
  s += strconv.Itoa(src) + "-"
  s += strconv.Itoa(dst)
  return s
}

func cleanpp(tag string, n int) {
  for i := 0; i < n; i++ {
    for j := 0; j < n; j++ {
      ij := pp(tag, i, j)
      os.Remove(ij)
    }
  }
}

func part(t *testing.T, tag string, npaxos int, p1 []int, p2 []int, p3 []int) {
  cleanpp(tag, npaxos)

  pa := [][]int{p1, p2, p3}
  for pi := 0; pi < len(pa); pi++ {
    p := pa[pi]
    for i := 0; i < len(p); i++ {
      for j := 0; j < len(p); j++ {
        ij := pp(tag, p[i], p[j])
        pj := port(tag, p[j])
				mu.Lock()
        os.Link(pj, ij)
				mu.Unlock()
      }
    }
  }
}

func TestPartition(t *testing.T) {
  runtime.GOMAXPROCS(4)

  tag := "partition"
  const npaxos = 5
  var pxa []*Paxos = make([]*Paxos, npaxos)
  defer cleanup(pxa)
  defer cleanpp(tag, npaxos)

  for i := 0; i < npaxos; i++ {
    var pxh []string = make([]string, npaxos)
    for j := 0; j < npaxos; j++ {
      if j == i {
        pxh[j] = port(tag, i)
      } else {
        pxh[j] = pp(tag, i, j)
      }
    }
    pxa[i] = Make(mencius, persist, pxh, i, nil)
  }
  defer part(t, tag, npaxos, []int{}, []int{}, []int{})

  seq := 0

  fmt.Printf("Test: No decision if partitioned ...\n")

  part(t, tag, npaxos, []int{0,2}, []int{1,3}, []int{4})
  pxa[1].Start(seq, 111)
  checkmax(t, pxa, seq, 0)
  
  fmt.Printf("  ... Passed\n")

  fmt.Printf("Test: Decision in majority partition ...\n")

  part(t, tag, npaxos, []int{0}, []int{1,2,3}, []int{4})
  time.Sleep(2 * time.Second)
  waitmajority(t, pxa, seq)

  fmt.Printf("  ... Passed\n")

  fmt.Printf("Test: All agree after full heal ...\n")

  pxa[0].Start(seq, 1000) // poke them
  pxa[4].Start(seq, 1004)
  part(t, tag, npaxos, []int{0,1,2,3,4}, []int{}, []int{})

  waitn(t, pxa, seq, npaxos)

  fmt.Printf("  ... Passed\n")

  fmt.Printf("Test: One peer switches partitions ...\n")

  for iters := 0; iters < 20; iters++ {
    seq++

    part(t, tag, npaxos, []int{0,1,2}, []int{3,4}, []int{})
    pxa[0].Start(seq, seq * 10)
    pxa[3].Start(seq, (seq * 10) + 1)
    waitmajority(t, pxa, seq)
    if ndecided(t, pxa, seq) > 3 {
      t.Fatalf("too many decided")
    }
    
    part(t, tag, npaxos, []int{0,1}, []int{2,3,4}, []int{})
    waitn(t, pxa, seq, npaxos)
  }

  fmt.Printf("  ... Passed\n")

  fmt.Printf("Test: One peer switches partitions, unreliable ...\n")


  for iters := 0; iters < 20; iters++ {
    seq++

    for i := 0; i < npaxos; i++ {
      pxa[i].unreliable = true
    }

    part(t, tag, npaxos, []int{0,1,2}, []int{3,4}, []int{})
    for i := 0; i < npaxos; i++ {
      pxa[i].Start(seq, (seq * 10) + i)
    }
    waitn(t, pxa, seq, 3)
    if ndecided(t, pxa, seq) > 3 {
      t.Fatalf("too many decided")
    }
    
    part(t, tag, npaxos, []int{0,1}, []int{2,3,4}, []int{})

    for i := 0; i < npaxos; i++ {
      pxa[i].unreliable = false
    }

    waitn(t, pxa, seq, 5)
  }

  fmt.Printf("  ... Passed\n")
}

func TestLots(t *testing.T) {

	if mencius {
		return
	}

	Lots(t, false)
}

func Lots(t *testing.T, crashes bool) {
  runtime.GOMAXPROCS(4)

	if crashes {
		fmt.Printf("Test: Many requests, changing partitions and random crashes ...\n")
	} else {
		fmt.Printf("Test: Many requests, changing partitions ...\n")
	}

	tag := ""
	if crashes {
		tag = "lotscrashes"
	} else {
		tag = "lots"
	}
  const npaxos = 5
  var pxa []*Paxos = make([]*Paxos, npaxos)
	defer cleanup(pxa)
  defer cleanpp(tag, npaxos)

  for i := 0; i < npaxos; i++ {
    var pxh []string = make([]string, npaxos)
    for j := 0; j < npaxos; j++ {
      if j == i {
        pxh[j] = port(tag, i)
      } else {
        pxh[j] = pp(tag, i, j)
      }
    }
    pxa[i] = Make(mencius, persist, pxh, i, nil)
    pxa[i].unreliable = true
  }

  defer part(t, tag, npaxos, []int{}, []int{}, []int{})

	doneChannel := make(chan bool)
	if crashes {
		go crashRandomly(false, pxa, doneChannel)
	}

  done := false

  // re-partition periodically
  ch1 := make(chan bool)
  go func() {
    defer func(){ ch1 <- true }()
    for done == false {
      var a [npaxos]int
      for i := 0; i < npaxos; i++ {
        a[i] = (rand.Int() % 3)
      }
      pa := make([][]int, 3)
      for i := 0; i < 3; i++ {
        pa[i] = make([]int, 0)
        for j := 0; j < npaxos; j++ {
          if a[j] == i {
            pa[i] = append(pa[i], j)
          }
        }
      }
      part(t, tag, npaxos, pa[0], pa[1], pa[2])
      time.Sleep(time.Duration(rand.Int63() % 200) * time.Millisecond)
    }
  }()

  seq := 0

  // periodically start a new instance
  ch2 := make(chan bool)
  go func () {
    defer func() { ch2 <- true } ()
    for done == false {
      // how many instances are in progress?
      nd := 0
      for i := 0; i < seq; i++ {
        if ndecided(t, pxa, i) == npaxos {
          nd++
        }
      }
      if seq - nd < 10 {
        for i := 0; i < npaxos; i++ {
					mu.Lock()
          pxa[i].Start(seq, rand.Int() % 10)
					mu.Unlock()
        }
        seq++
      }
      time.Sleep(time.Duration(rand.Int63() % 300) * time.Millisecond)
    }
  }()

  // periodically check that decisions are consistent
  ch3 := make(chan bool)
  go func() {
    defer func() { ch3 <- true }()
    for done == false {
      for i := 0; i < seq; i++ {
        ndecided(t, pxa, i)
      }
      time.Sleep(time.Duration(rand.Int63() % 300) * time.Millisecond)
    }
  }()

  time.Sleep(20 * time.Second)
  done = true
  <- ch1
  <- ch2
  <- ch3

	if crashes {
		doneChannel <- true
		<-doneChannel
	}

  // repair, then check that all instances decided.
  for i := 0; i < npaxos; i++ {
    pxa[i].unreliable = false
  }
  part(t, tag, npaxos, []int{0,1,2,3,4}, []int{}, []int{})
  time.Sleep(5 * time.Second)

  for i := 0; i < seq; i++ {
    waitmajority(t, pxa, i)
  }

  fmt.Printf("  ... Passed\n")
}

func TestBasicCrashes(t *testing.T) {
  runtime.GOMAXPROCS(4)

  const npaxos = 3
  var pxa []*Paxos = make([]*Paxos, npaxos)
  var pxh []string = make([]string, npaxos)
  defer cleanup(pxa)
	
  for i := 0; i < npaxos; i++ {
    pxh[i] = port("basiccrash", i)
  }
  for i := 0; i < npaxos; i++ {
    pxa[i] = Make(mencius, persist, pxh, i, nil)
  }

	seq := 0
	
  fmt.Printf("Test: Crashes after agreement ...\n")

	value := "baby, don't die on me"
  pxa[0].Start(seq, value)
  waitn(t, pxa, seq, npaxos)
	pxa[0].Kill(true)
	pxa[1].Kill(false)
	pxa[0] = Make(mencius, persist, pxh, 0, nil)
	pxa[1] = Make(mencius, persist, pxh, 1, nil)
	pxa[0].Start(seq, "i'm wrong")
	pxa[1].Start(seq, "i'm wrong")
  waitn(t, pxa, seq, npaxos)

	_,v1 := pxa[0].Status(0)	
  if v1 != value {
    t.Fatalf("incorrect value decided, found %v and not %v", v1, value)
  }

  fmt.Printf("  ... Passed\n")

  fmt.Printf("Test: Crashes before agreement ...\n")

	seq += 1
  pxa[0].Start(seq, "i believe in trust")
	time.Sleep(time.Millisecond)
	pxa[0].Kill(true)
	pxa[0] = Make(mencius, persist, pxh, 0, nil)
  pxa[0].Start(seq, "i'm different")
  waitn(t, pxa, seq, npaxos)

  fmt.Printf("  ... Passed\n")

  fmt.Printf("Test: Multiple peers crash repeatedly before agreement ...\n")

	seq += 1
  pxa[0].Start(seq, "late night, i go to sleep")
	pxa[1].Kill(false)
	pxa[2].Kill(true)
	pxa[1] = Make(mencius, persist, pxh, 1, nil)
  pxa[1].Start(seq, "1")
	pxa[1].Kill(true)
	pxa[1] = Make(mencius, persist, pxh, 1, nil)
  pxa[1].Start(seq, "2")
	pxa[1].Kill(false)
	pxa[2] = Make(mencius, persist, pxh, 2, nil)
	pxa[1] = Make(mencius, persist, pxh, 1, nil)
  pxa[1].Start(seq, "2")
  pxa[2].Start(seq, "3")
  waitn(t, pxa, seq, npaxos)

  fmt.Printf("  ... Passed\n")

  fmt.Printf("Test: Everyone crashes after agreement (one disk lost) ...\n")

	seq += 1
	value = "wondering what is real"
  pxa[0].Start(seq, value)
  waitn(t, pxa, seq, npaxos)
	pxa[0].Kill(false)
	pxa[1].Kill(true)
	pxa[2].Kill(true)
	pxa[0] = Make(mencius, persist, pxh, 0, nil)
	pxa[1] = Make(mencius, persist, pxh, 1, nil)
	pxa[2] = Make(mencius, persist, pxh, 2, nil)
  pxa[0].Start(seq, "i'm wrong")
  //pxa[1].Start(seq, "i'm wrong")
  //pxa[2].Start(seq, "i'm wrong")
	waitn(t, pxa, seq, npaxos)

	_,v1 = pxa[0].Status(seq)	
  if v1 != value {
    t.Fatalf("incorrect value decided found %v ,  instead of %v, in slot %v", v1, value, seq)
  }

  fmt.Printf("  ... Passed\n")
}

func TestDeleting(t *testing.T) {
  runtime.GOMAXPROCS(4)

  const npaxos = 5
  var pxa []*Paxos = make([]*Paxos, npaxos)
  var pxh []string = make([]string, npaxos)

  for i := 0; i < npaxos; i++ {
    pxh[i] = port("del", i)
  }
  for i := 0; i < npaxos; i++ {
    pxa[i] = Make(mencius, persist, pxh, i, nil)
  }

  defer cleanup(pxa)

  fmt.Printf("Test: No record of instances from old tests ...\n")

	value := rand.Int() % 10
	newVal := rand.Intn(2) < 1
	if newVal {
		safeStart(pxa, 0, 0, value)
		waitn(t, pxa, 0, npaxos)
	}

	decided,_ := pxa[0].Status(0)
  if !newVal && decided {
    t.Fatalf("incorrect value decided")
  }

  fmt.Printf("  ... Passed\n")
}


func TestDead(t *testing.T) {
  runtime.GOMAXPROCS(4)

  const npaxos = 5
  var pxa []*Paxos = make([]*Paxos, npaxos)
  var pxh []string = make([]string, npaxos)
  defer cleanup(pxa)

  for i := 0; i < npaxos; i++ {
    pxh[i] = port("dead", i)
  }
  for i := 0; i < npaxos; i++ {
    pxa[i] = Make(mencius, persist, pxh, i, nil)
  }

  fmt.Printf("Test: Dead proposer ...\n")

  pxa[0].Start(0, "hello")
  waitn(t, pxa, 0, npaxos)

	pxa[0].Kill(true)
	pxa[npaxos - 1].Kill(true)

  pxa[1].Start(1, "goodbye")
  waitmajority(t, pxa, 1)
  time.Sleep(1 * time.Second)
  if ndecided(t, pxa, 1) != npaxos - 2 {
    t.Fatalf("a dead peer heard about a decision")
  }

	pxa[0] = Make(mencius, persist, pxh, 0, nil)
  pxa[0].Start(1, "xxx")
  waitn(t, pxa, 1, npaxos-1)
  time.Sleep(1 * time.Second)
  if ndecided(t, pxa, 1) != npaxos - 1 {
    t.Fatalf("a dead peer heard about a decision")
  }

	pxa[npaxos-1] = Make(mencius, persist, pxh, npaxos-1, nil)
  pxa[npaxos-1].Start(1, "yyy")
  waitn(t, pxa, 1, npaxos)

  fmt.Printf("  ... Passed\n")
}

//
// does paxos forgetting actually free the memory with crashes?
//
func TestForgetMemCrashes(t *testing.T) {
	ForgetMem(t, true)
}

//
// many agreements (with failures)
//
func TestManyCrashes(t *testing.T) {
	Many(t, true)
}

//
// paxos reaches agreement, two die, and two new processes are created
// 
func TestOldCrashes(t *testing.T) {
  runtime.GOMAXPROCS(4)

  fmt.Printf("Test: Minority proposal ignored after crashes ...\n")

  const npaxos = 5
  var pxa []*Paxos = make([]*Paxos, npaxos)
  var pxh []string = make([]string, npaxos)
  defer cleanup(pxa)

  for i := 0; i < npaxos; i++ {
    pxh[i] = port("oldcrashes", i)
  }

  pxa[1] = Make(mencius, persist, pxh, 1, nil)
  pxa[2] = Make(mencius, persist, pxh, 2, nil)
  pxa[3] = Make(mencius, persist, pxh, 3, nil)
  pxa[1].Start(1, 111)

  waitmajority(t, pxa, 1)

	pxa[1].Kill(true)
	pxa[2].Kill(false)
	pxa[3].Kill(false)
	
  pxa[0] = Make(mencius, persist, pxh, 0, nil)
  pxa[1] = Make(mencius, persist, pxh, 1, nil)
	pxa[4] = Make(mencius, persist, pxh, 4, nil)
  pxa[0].Start(1, 222)

  waitmajority(t, pxa, 1)

	decided, v1 := pxa[0].Status(1)
	if !decided {
			t.Fatalf("zombie processes serving requests! %v", v1)
	}
	if v1 != 111 {
		t.Fatalf("wrong value decided!")
	}
	
  fmt.Printf("  ... Passed\n")
}

//
// many agreements, with unreliable RPC and crashes
//
func TestManyUnreliableCrashes(t *testing.T) {
	ManyUnreliable(t, true)
}

func TestLotsCrashes(t *testing.T) {
	Lots(t, true)
}

func TestMenciusRPC(t *testing.T) {
  runtime.GOMAXPROCS(4)

  const npaxos = 3
  var pxa []*Paxos = make([]*Paxos, npaxos)
  var pxh []string = make([]string, npaxos)
  defer cleanup(pxa)

  for i := 0; i < npaxos; i++ {
    pxh[i] = port("basic", i)
  }
  for i := 0; i < npaxos; i++ {
    pxa[i] = Make(mencius, persist, pxh, i, nil)
  }
 
  //init done channels
  var chans [npaxos]chan bool
  //fmt.Println(len(chans))
  for i := range chans {
    chans[i] = make(chan bool)
  } 

  // launch proposers

  go func() {
    k := 0
    defer func() { chans[k] <- true }()
    s := 0
    for j := 0; j < 100; {
      //fmt.Println("loop")
      s = pxa[k].GetNextSlot()
      pxa[k].Start(s, j)
      waitn(t, pxa, s, npaxos)
      _, v1 := pxa[k].Status(s)
      if v1 == j {
        //fmt.Println("win", j)
        j++
      }
    }
  }()
  
  go func() {
    k := 1
    defer func() { chans[k] <- true}()
    s := 0
    for j := 0; j < 100;  {
      s = pxa[k].GetNextSlot()
      pxa[k].Start(s, j)
      waitn(t, pxa, s, npaxos)
      _, v1 := pxa[k].Status(s)
      if v1 == j {
        j++
      }
    }
  }()

  go func() {
    k := 2
    defer func() { chans[k] <- true}()
    s := 0
    for j := 0; j < 100;  {
      s = pxa[k].GetNextSlot()
      pxa[k].Start(s, j)
      waitn(t, pxa, s, npaxos)
      _, v1 := pxa[k].Status(s)
      if v1 == j {
        j++
      }
    }
  }()

  for i := 0; i < npaxos; i++ {
    <- chans[i]
  }

  // get rppc counts
  total := 0
  for j := 0; j < npaxos; j++ {
    total += pxa[j].rpcCount
  }
  fmt.Printf("RPC Count : %d\n", total)
  fmt.Printf("  ... Passed\n")
}
